# packages/static-pages/tcl/static-pages-procs.tcl
ad_library {
    Utilities for static pages.

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-01-22
    @cvs-id $Id: static-pages-procs.tcl,v 1.8 2001/08/27 23:19:05 andy Exp $
}


ad_proc -public sp_sync_cr_with_filesystem {
    {
        -file_add_proc ""
        -file_change_proc ""
        -file_unchanged_proc ""
        -folder_add_proc ""
        -folder_unchanged_proc ""
        -package_id ""
    }
    fs_root
    root_folder_id
    {
       static_page_regexp {\.(html|htm|pdf)$}
    }
} {
    Synchronize the content repository with the file system.
    This creates entries in sp_folders and static_pages, so the static_page
    functions must be used to delete entries.

    @param file_add_proc The name of a Tcl proc to be called for each file
                         added.  The full file path and the page_id will be
                         passed to it.

    @param file_change_proc The name of a Tcl proc to be called for each file
                            changed in the database.

    @param file_unchanged_proc The name of a Tcl proc to be called for each file
                            unchanged in the database.

    @param folder_add_proc The name of a Tcl proc to be called for each folder
                           added.  The full file path and the folder_id will be
                           passed to it.

    @param folder_unchanged_proc The name of a Tcl proc to be called for each folder
                                 unchanged in the database.

    @param fs_root The starting path in the filesystem.  Files below this point will 
                   be scanned.

    @param root_folder_id The id of the root folder in the static-pages system (and in 
                          the content repository) obtained from 
                          <code>static_page.get_root_folder</code>.

    @param static_page_regexp A regexp to identify static pages.

    @param package_id Optionally, the package id of the Static Pages
    instance.  If not specified, determined from ad_conn.

    @author Andrew Piskorski (atp@piskorski.com)
    @creation-date 2001/08/27
} {
   if { [empty_string_p $package_id] } {
      set package_id [ad_conn package_id]
   }

   if { [catch { set return_val [sp_sync_cr_with_filesystem_internal \
           -file_add_proc          $file_add_proc         \
           -file_change_proc       $file_change_proc      \
           -file_unchanged_proc    $file_unchanged_proc   \
           -folder_add_proc        $folder_add_proc       \
           -folder_unchanged_proc  $folder_unchanged_proc \
           -package_id             $package_id            \
           -stack_depth  2 \
           {return_mesg}  $fs_root  $root_folder_id  $static_page_regexp ]
   } result] } {
      # We caught an unexpected error, so clean up the mutex, and then
      # re-throw the exact same error:

      sp_sync_cr_with_filesystem_unlock $package_id

      global errorInfo
      error $result $errorInfo
   } else {
      return $return_mesg
   }
}


ad_proc -private sp_sync_cr_with_filesystem_unlock {package_id} {
   Unlocks the sp_sync_cr_with_filesystem_times variable - use upon
   abnormal termination of the sp_sync_cr_with_filesystem_internal
   stuff.  We have it as a separate proc here to make it convenient to
   call from within multiple different procedures.

   @author Andrew Piskorski (atp@piskorski.com)
   @creation-date 2001/08/27
} {
   ns_share sp_sync_cr_with_filesystem_times
   ns_share sp_sync_cr_with_filesystem_mutex

   ns_mutex lock $sp_sync_cr_with_filesystem_mutex
   set sp_sync_cr_with_filesystem_times($package_id) {}
   ns_mutex unlock $sp_sync_cr_with_filesystem_mutex
}


ad_proc -private sp_sync_cr_with_filesystem_internal { 
    {
        -file_add_proc ""
        -file_change_proc ""
        -file_unchanged_proc ""
        -folder_add_proc ""
        -folder_unchanged_proc ""
        -package_id ""
        -stack_depth 1
    }
    return_mesg_var
    fs_root
    root_folder_id
    {
       static_page_regexp {\.(html|htm|pdf)$}
    }
} {
   This procedure was originally named sp_sync_cr_with_filesystem
   procedure, but has been renamed and modified so that it can be
   wrapped inside the new sp_sync_cr_with_filesystem, to support the
   mutex locking.
   <p>
   We wrap it because at the end of this proc, we must set
   sp_sync_cr_with_filesystem_times($package_id) back to empty string.
   But if we hit some random untrapped error partway through, we'll
   never get there.  Therefore, we wrap this proc inside another, and
   have the wrapper proc catch any errors thrown by this proc, set the
   var back to empty string, then re-throw the error.
   <p>
   This procedure takes the exact same arguments as its
   sp_sync_cr_with_filesystem wrapper proc, except for the addition of
   return_mesg_var.
   <p>
   You should <em>never</em> call this procedure, except from
   sp_sync_cr_with_filesystem.

   @param return_mesg_var Name of variable in which to return text
   message, for presentation on a web page to the user.

   @author Brandoch Calef (bcalef@arsdigita.com)
   @author Andrew Piskorski (atp@piskorski.com)
   @creation-date 2001-02-07
} {
    if { [empty_string_p $package_id] } {
        set package_id [ad_conn package_id]
    }
    upvar $return_mesg_var return_mesg
    set return_mesg {}


    # Make sure that only 1 copy of this proc per package instance
    # ever runs at once:

    ns_share sp_sync_cr_with_filesystem_times
    ns_share sp_sync_cr_with_filesystem_mutex

    ns_mutex lock $sp_sync_cr_with_filesystem_mutex
    if { -1 == [lsearch [array names sp_sync_cr_with_filesystem_times] $package_id] } {
       # The package_id isn't in the array yet at all, so another copy
       # is not running.
       set sp_sync_cr_with_filesystem_times($package_id) [ns_time]
    } elseif { [empty_string_p $sp_sync_cr_with_filesystem_times($package_id)] } {
       # We're ok, no other copy is running.
       set sp_sync_cr_with_filesystem_times($package_id) [ns_time]
    } else {
       # Another copy is running.
       set other_start_time $sp_sync_cr_with_filesystem_times($package_id)
       ns_mutex unlock $sp_sync_cr_with_filesystem_mutex

       set other_time_pretty [ns_httptime $other_start_time]
       set time_diff [expr [ns_time] -  $other_start_time]

       set mesg "sp_sync_cr_with_filesystem: Already running. sp_sync_cr_with_filesystem_times($package_id) == $other_time_pretty, $time_diff seconds ago."
       ns_log Warning $mesg

       set return_mesg "Another copy of this procedure is already running for
       this package instance.  It started running $time_diff seconds
       ago, at $other_time_pretty.  Only one copy may run at a time.
       Please wait and then try again."

       return 0

       # Instead of [ns_httptime [ns_time]], could also use [clock format [clock seconds]].
    }
    ns_mutex unlock $sp_sync_cr_with_filesystem_mutex

    # Up to here in this proc had better be error free - if not, our
    # whole mutex and error catching scheme will probably crash and
    # burn!  --atp@piskorski.com, 2001/08/27 18:37 EDT


    set sync_session_id [db_nextval sp_session_id_seq]

    set fs_trimmed [string trimright $fs_root "/"]
    set fs_trimmed_length [string length $fs_trimmed]

    foreach file [ad_find_all_files $fs_root] {
        if { [regexp -nocase $static_page_regexp $file match] } {
            # Chop the starting path off of the full pathname and split it up:
            set path [split [string range $file $fs_trimmed_length end] "/"]
            # Throw away the first entry (empty) and the last entry (which is the filename):
            set path [lrange $path 1 [expr [llength $path]-2]]

            set cumulative_path ""
            set parent_folder_id $root_folder_id
            foreach directory $path {
                append cumulative_path "$directory/"
                if (![info exists path_exists($cumulative_path)]) {
                    # check db
                    set folder_id [db_string get_folder_id {
                        select nvl(content_item.get_id(:cumulative_path,:root_folder_id),0)
                        from dual
                    }]
                    # If the folder doesn't exist, create it.
                    if { $folder_id == 0} {
                        set folder_id [db_exec_plsql create_new_folder {
                            begin
                                    :1 := static_page.new_folder (
                                            name        => :directory,
                                            label       => :directory,
                                            parent_id   => :parent_folder_id,
                                            description => 'Static pages folder'
                                    );
                            end;
                        }]
                        if { [string length $folder_add_proc] > 0 } {
                            uplevel $stack_depth "$folder_add_proc $cumulative_path $folder_id"
                        }
                    } else {
                        if { [string length $folder_unchanged_proc] > 0 } {
                            uplevel $stack_depth "$folder_unchanged_proc $cumulative_path $folder_id"
                        }
                    }
                    set path_exists($cumulative_path) $folder_id
                    db_dml insert_path {
                        insert into sp_extant_folders (session_id,folder_id)
                        values (:sync_session_id,:folder_id)
                    }
                } else {
                    set folder_id $path_exists($cumulative_path)
                }
                set parent_folder_id $folder_id
            }

            # If the file is already in the db:
            #    Fetch it from the db and load the file from the filesystem
            #    If they differ:
            #        Insert the filesystem version into the db.
            # If the file isn't in the db:
            #    Insert it.

            set fs_filename $file
            if [db_0or1row check_db_for_page {
                select static_page_id from static_pages
                where filename = :fs_filename
            }] {
                db_1row get_db_page {
                    select content as file_from_db from cr_revisions
                    where revision_id = content_item.get_live_revision(:static_page_id)
                }
                if { [catch {
                    set fp [open $file r]
                    set file_from_fs [read $fp]
                    close $fp
                } errmsg]} {
                    ad_return_error "Error reading file" \
                            "This error was encountered while reading $file: $errmsg"
                    sp_sync_cr_with_filesystem_unlock $package_id
                    set return_mesg "Error reading file."
                    return 0
                }
                if { $file_from_fs != $file_from_db } {
                    db_dml update_db_file {
                        update cr_revisions set content = empty_blob()
                        where revision_id = content_item.get_live_revision(:static_page_id)
                        returning content into :1
                    } -blob_files [list $file]
                        if { [string length $file_change_proc] > 0 } {
                            uplevel $stack_depth "$file_change_proc $file $static_page_id"
                        }
                } else {
                    if { [string length $file_unchanged_proc] > 0 } {
                        uplevel $stack_depth "$file_unchanged_proc $file $static_page_id"
                    }
                }
                db_dml insert_file {
                    insert into sp_extant_files (session_id,static_page_id)
                    values (:sync_session_id,:static_page_id)
                }
            } else {
                # Try to extract a title:
                if { [catch {
                    set fp [open $file r]
                    set file_contents [read $fp]
                    close $fp
                } errmsg]} {
                    ad_return_error "Error reading file" \
                            "This error was encountered while reading $file: $errmsg"
                    sp_sync_cr_with_filesystem_unlock $package_id
                    set return_mesg "Error reading file."
                    return 0
                }

                # TODO:  This is very HTML specific:  --atp@piskorski.com, 2001/08/13 21:58 EDT
                if { ![regexp -nocase {<title.*?>(.+?)</title} $file_contents match page_title] } {
                    regexp {[^/]*$} $file page_title
                }

                # Insert into the db:
                # the Oracle driver apparently doesn't support passing BLOBs to
                # PL/SQL (or passing a BLOB from a file into a PL/SQL function),
                # so the PL/SQL call and the BLOB loading must be performed
                # seperately.  This is simple (get item_id from static_page.new(),
                # then update cr_revisions to insert the blob) but involved direct
                # manipulation of the cr_revisions table.

                # TODO: Note that fs_maybe_create_new_mime_type is
                # defined in the File Storage package only.
                set mime_type [fs_maybe_create_new_mime_type $file]

                # If you run two copies of sp_sync_cr_with_filesystem
                # at once, you CAN get "ORA-00001: unique constraint
                # (CR_ITEMS_UNIQUE_NAME) violated" errors here when
                # calling static_page.new - thus the addition of mutex
                # locking.  --atp@piskorski.com, 2001/08/27 01:20 EDT

                set static_page_id [db_exec_plsql do_sp_new {
                    begin
                      :1 := static_page.new(
                        filename  => :file,
                        title     => :page_title,
                        folder_id => :parent_folder_id,
                        mime_type => :mime_type
                      );
                    end;
                }]
                # Check if -blobs [list $file_contents] would be faster:
                db_dml insert_file_contents {
                    update cr_revisions set content = empty_blob()
                    where revision_id = content_item.get_live_revision(:static_page_id)
                    returning content into :1
                } -blob_files [list $file]
                if { [string length $file_add_proc] > 0 } {
                    uplevel $stack_depth "$file_add_proc $file $static_page_id"
                }
                db_dml insert_file {
                    insert into sp_extant_files (session_id,static_page_id)
                    values (:sync_session_id,:static_page_id)
                }
            }
        }
    }

    # TODO: This is very wrong.  Should NEVER just delete all content!
    # See also: http://openacs.org/bboard/q-and-a-fetch-msg.tcl?msg_id=0002U2
    # --atp@piskorski.com, 2001/08/13 15:07 EDT

    # Clean up any files that are in the db but no longer in the filesystem:

    # TODO: Why are we doing these two deletes after calling
    # static_page.delete_stale_items?  --atp@piskorski.com, 2001/08/23
    # 02:20 EDT

    db_exec_plsql delete_old_files {
        begin
            static_page.delete_stale_items(:sync_session_id,:package_id);

            delete from sp_extant_folders where session_id = :sync_session_id;
            delete from sp_extant_files where session_id = :sync_session_id;
        end;
    }

    sp_sync_cr_with_filesystem_unlock $package_id
    set return_mesg "Done."
    return 0
}


ad_proc -public sp_root_folder_id { package_id } {
    Returns the id of the root folder associated with package_id,
    creating one if necessary.

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-02-23
} {
    # We must use db_exec_plsql rather than simply selecting from dual
    # because static_page.get_root_folder will do DML (to create the
    # folder) if it can't find a root folder.
    return [db_exec_plsql get_root_folder_id {
        begin
            :1 := static_page.get_root_folder(:package_id);
        end;
    }]
}


ad_proc -public sp_change_matching_permissions {
    root_folder_id
    contained_string
    grant_or_revoke
} {
    Grant or revoke permissions on all files below root_folder_id
    whose filenames contain contained_string.

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-02-23
} {
    if { $grant_or_revoke != "grant" && $grant_or_revoke != "revoke" } {
        ns_log Warning "sp_change_matching_permissions called with grant_or_revoke = $grant_or_revoke"
        return
    }

    db_exec_plsql grant_or_revoke_matching_permissions "
            begin
            for file_row in (
                    select static_page_id from static_pages
                    where folder_id in (
                            select folder_id from sp_folders
                            start with folder_id = :root_folder_id
                            connect by parent_id = prior folder_id)
                    and filename like '%${contained_string}%'
            ) loop
                    acs_permission.${grant_or_revoke}_permission(
                            object_id => file_row.static_page_id,
                            grantee_id => acs.magic_object_id('the_public'),
                            privilege => 'general_comments_create'
                    );
            end loop;
            end;
    "
}


ad_proc -public sp_change_matching_display {
    root_folder_id
    contained_string
    show_full_comments_p
} {
    Set all files below root_folder_id whose filenames contain 
    contained_string to have comments either shown (full contents of
    comments are displayed on the page) or summarized (title line of
    comments are listed).

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-02-23
} {
    if { $show_full_comments_p != "t" && $show_full_comments_p != "f" } {
        ns_log Warning "sp_change_matching_permissions called with show_full_comments_p = $show_full_comments_p"
        return
    }

    db_foreach matching_static_page "
        select static_page_id from static_pages
             where folder_id in (
                     select folder_id from sp_folders
                     start with folder_id = :root_folder_id
                     connect by parent_id = prior folder_id)
             and filename like '%${contained_string}%'
    " {
        sp_flush_page $static_page_id
    }   

    db_dml show_or_summarize_comments_matching "
            update static_pages set show_comments_p = :show_full_comments_p 
                where static_page_id in (
                    select static_page_id from static_pages
                    where folder_id in (
                            select folder_id from sp_folders
                            start with folder_id = :root_folder_id
                            connect by parent_id = prior folder_id)
                    and filename like '%${contained_string}%'
                )
    "
}


ad_proc -private sp_get_page_info_query { page_id } {
    Returns a SQL query to get the page title and comment display policy.

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-02-23
} {
    return "select '{'||content_item.get_title($page_id)||'} '||decode(show_comments_p,'t',1,0) from static_pages where static_page_id = $page_id"
}


ad_proc -public sp_flush_page { page_id } {
    Flushes the cache entry for a static page.  This should be done whenever the
    page title or show_comment_p setting change.

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-02-23
} {
    util_memoize_flush [list db_string get_page_info [sp_get_page_info_query $page_id]]
}


ad_proc -public sp_package_key_is {} {
   Simply returns the package key string for this package.

   @author Andrew Piskorski (atp@piskorski.com)
   @creation-date 2001/08/26
} {
   # TODO: Might want to have this pull and cache the actual key from
   # the database.

   return {static-pages}
}


ad_proc -public sp_serve_html_page { } {
    Registered proc to serve up static pages.

    @author Brandoch Calef (bcalef@arsdigita.com)
    @creation-date 2001-01-23
} {
    set filename [ad_conn file]

    # Using apm_package_id_from_key only works if static-pages is a
    # singleton package, so I removed that.  --atp@piskorski.com,
    # 2001/08/26 22:37 EDT

    set query "
      select static_page_id
      from static_pages sp, sp_folders f
      where sp.filename = '$filename'
        and sp.folder_id = f.folder_id
        -- Only want pages from the Static Pages package.
        and f.package_id in (
          select package_id  from apm_packages
          where package_key = '[sp_package_key_is]' )
        -- If the same page is in more than one instance of
        -- Static Pages for some reason, we only want one of
        -- them, and we don't care which.
        and rownum <= 1
    "
    set page_id [util_memoize [list db_string search_page $query -default -1]]

    # If the page is in the db, serve it carefully; otherwise just dump it out.
    if { $page_id >= 0 } {
        set page_info [util_memoize [list db_string get_page_info [sp_get_page_info_query $page_id]]]

        # TODO: Below, what if we only allow registered users to make
        # comments?  Or some smaller group of users?  What then?
        # --atp@piskorski.com, 2001/08/22 23:09 EDT

        # We only show the link here if the_public has 
        # general_comments_create privilege on the page.  Why the_public
        # rather than the current user?  Because we don't want admins to
        # be seeing "Add a comment" links on non-commentable pages.
        #
        set comment_link ""
        if { [ad_permission_p -user_id [acs_magic_object the_public] $page_id general_comments_create] } {
            append comment_link "<center>
            [general_comments_create_link -object_name [lindex $page_info 0] $page_id [ad_conn url]]
            </center>"
        }
        append comment_link "[general_comments_get_comments -print_content_p [lindex $page_info 1] $page_id [ad_conn url]]"

        if { [catch {
            set fp [open $filename r]
            set file_contents [read $fp]
            close $fp
        } errmsg] } {
            ad_return_error "Error reading file" \
                    "This error was encountered while reading $filename: $errmsg"
        }

        # Tcl needs a case-insensitive [string first] function.
        #
        set body_close [string first "</body" [string tolower $file_contents]]
        if { $body_close >= 0 } {
            doc_return 200 text/html "[string range $file_contents 0 [expr $body_close-1]]${comment_link}[string range $file_contents $body_close end]"
        } else {
            doc_return 200 text/html "${file_contents}$comment_link"
        }
    } else {
        ns_returnfile 200 text/html $filename
    }
}


# Register the handler for each static page file extension.
rp_register_extension_handler html sp_serve_html_page
rp_register_extension_handler htm sp_serve_html_page
rp_register_extension_handler HTML sp_serve_html_page
rp_register_extension_handler HTM sp_serve_html_page

# TODO:  Add a PDF extension handler?

